/*
 * Copyright (c) AppDynamics, Inc., and its affiliates
 * 2016
 * All Rights Reserved
 * THIS IS UNPUBLISHED PROPRIETARY CODE OF APPDYNAMICS, INC.
 * The copyright notice above does not evidence any actual or intended publication of such source code
 */
'use strict';
function GrpcExitProbe(agent) {
  this.agent = agent;
}

exports.GrpcExitProbe = GrpcExitProbe;
GrpcExitProbe.prototype.init = function () {
};

GrpcExitProbe.prototype.attach = function(obj) {
  var self = this;
  var proxy = self.agent.proxy;
  var address;
  var serviceClients = [];
  self.module = obj;
  
  function populateMethodsList(methodDefinitions){
    var methodTypes = {}; //Map containing the names of the methods defined in gRPC Server as keys, and their respective method types as values.
    var methodType;
    for (var method in methodDefinitions){
      if (methodDefinitions[method].requestStream) {
        if (methodDefinitions[method].responseStream) {
          methodType = 'bidi';
        } else {
          methodType = 'client_stream';
        }
      } else {
        if (methodDefinitions[method].responseStream) {
          methodType = 'server_stream';
        } else {
          methodType = 'unary';
        }
      }
      methodTypes[methodDefinitions[method].originalName] = methodType; 
    }
    return methodTypes;
  }

  function findServiceClients(servicePack) {
    let packageName = Object.keys(servicePack)[0];
    let serviceClientList = [];
    let servicePackage = servicePack[packageName];
    for (var serviceName in servicePackage) { 
      var service = servicePackage[serviceName];
      if (typeof(service) == 'function') {
        serviceClientList.push(service);
      }
    }
    return serviceClientList;
  }

  //Create a function that invokes the proxy.around(serviceClient.prototype)
  function invokeInterception(serviceClient, method, serviceName, methodTypes) {
    var threadId = self.agent.thread.current();
    proxy.around(serviceClient, method, function(obj, args, locals) { 
      var supportedProperties = {
        'SERVICE': serviceName,
        'VENDOR': 'gRPC',
        'OPERATION': method,
        'SOAP_ACTION': methodTypes[method],
        'URL': address
      };
      self.createExitCall(supportedProperties, method, args, locals);

      if (locals.exitCall) {
        var correlationHeaderValue = self.agent.backendConnector.getCorrelationHeader(locals.exitCall);
        if (correlationHeaderValue) {
          var argsArray = Array.prototype.slice.call(args);

          self.injectCorrelationHeader(argsArray, methodTypes[method], correlationHeaderValue);
          return argsArray;
        }
      }
    }, after, false, threadId);
  }

  proxy.after(obj.Client, 'call', function(obj, args){
    address = args[1];
  },false);

  proxy.after(obj, 'loadPackageDefinition' , function(obj,args, ret){
    //Retrieve gRPC address
    serviceClients = findServiceClients(ret);
    for (var i = 0; i < serviceClients.length; i++){
      let pathToFunction = Object.keys(serviceClients[i].prototype.$method_names)[0]; //Retrieve the path to the first function defined in a gRPC Server.
      let methodTypes = populateMethodsList(serviceClients[i].prototype.$method_definitions);
      let serviceName = pathToFunction.split('/')[1]; //From path to first function retrieve the service name for the specific ServiceClient
      for(var method in methodTypes){ 
        invokeInterception(serviceClients[i].prototype, method, serviceName, methodTypes);
      }
    }
  }, false); 

  function after(obj, args, ret, locals) {
    let finished = false;
    if(ret.constructor.name == 'ClientReadableStream' || 'ClientDuplexStream'){
      ret.on('status', () => {
        finished = true;
        self.complete(null, locals);
      });
      ret.on('error', () => {
        if(!finished){
          finished = true;
          self.complete(ret, locals);
        }
      });
    }
    if (locals.methodHasCb) {
      return;
    }
    if (!ret || !ret.__appdynamicsIsPromiseResult__) {
      self.complete(null, locals);
    } else if (ret.error) { 
      self.complete(ret.error, locals);
    } else {
      self.complete(null, locals);
    }
  }

};

GrpcExitProbe.prototype.injectCorrelationHeader = function(args, type, corrHeader) {
  var self = this;
  // This finds an instance of Metadata among the arguments.
  var metadataIndex = args.findIndex((arg) => {
    return (arg &&
            typeof arg === 'object' &&
            arg._internal_repr &&
            typeof arg.getMap === 'function'
    );
  });

  var metadata;
  if (metadataIndex == -1) {
    metadata = new self.module.Metadata();
    if (type == 'unary' || type == 'server_stream') {
      if (args.length == 0) {
        // No argument (for the gRPC call) was provided, so we will have to
        // provide one, since metadata cannot be the first argument.
        // The internal representation of argument defaults to undefined
        // in its non-presence.
        args.push(undefined);
      }
      metadataIndex = 1;
    } else {
      metadataIndex = 0;
    }
    args.splice(metadataIndex, 0, metadata);
  } else {
    metadata = args[metadataIndex];
  }
  metadata.set(self.agent.correlation.HEADER_NAME, corrHeader.toString());
};

GrpcExitProbe.prototype.createExitCall = function(supportedProperties, command, commandArgs, locals) {
  var self = this, profiler = self.agent.profiler;
  locals.time = profiler.time();
  locals.exitCall = profiler.createExitCall(locals.time, {
    exitType: 'WEB_SERVICE',
    supportedProperties: supportedProperties,
    command: command,
    commandArgs: commandArgs,
    stackTrace: profiler.stackTrace()
  });
  if (!locals.exitCall) return;
  if(supportedProperties['SOAP_ACTION'] == 'client_stream' || supportedProperties['SOAP_ACTION'] == 'unary') {
    locals.methodHasCb = self.agent.proxy.callback(commandArgs, -1, function(obj, args) {
      self.complete(args, locals);
    });
  }
};

GrpcExitProbe.prototype.complete = function(err, locals) {
  if (!locals.exitCall) return;
  if (!locals.time.done()) return;
  var error = this.agent.proxy.getErrorObject(err);
  this.agent.profiler.addExitCall(locals.time, locals.exitCall, error);
};
